import torch
import warnings
import numpy as np
import pandas as pd
import torch.nn as nn
import matplotlib.pyplot as plt

from captum.attr import Saliency
from finol.evaluation_layer.MetricCaculator import MetricCaculator
from finol.evaluation_layer.DistillerSelector import DistillerSelector
from finol.utils import load_config, portfolio_selection, actual_portfolio_selection

plt.style.use("seaborn-paper")
# plt.rcParams["font.family"] = "Microsoft YaHei"
warnings.filterwarnings("ignore")
warnings.simplefilter(action="ignore", category=FutureWarning)


class ModifiedModel(nn.Module):
    def __init__(self, model):
        super(ModifiedModel, self).__init__()
        self.base_model = model

    def forward(self, x):
        x = self.base_model(x)
        x = actual_portfolio_selection(x)
        return x


class EconomicDistiller:
    def __init__(self, load_dataset_output, train_model_output):
        self.config = load_config()
        self.load_dataset_output = load_dataset_output
        self.train_model_output = train_model_output
        self.logdir = train_model_output["logdir"]

        self.test_loader = load_dataset_output["test_loader"]
        self.NUM_TEST_PERIODS = load_dataset_output["NUM_TEST_PERIODS"]
        self.NUM_ASSETS = load_dataset_output["NUM_ASSETS"]
        self.NUM_FEATURES_ORIGINAL = load_dataset_output["NUM_FEATURES_ORIGINAL"]
        self.DETAILED_NUM_FEATURES = load_dataset_output["DETAILED_NUM_FEATURES"]
        self.WINDOW_SIZE = load_dataset_output["WINDOW_SIZE"]
        self.OVERALL_FEATURE_LIST = load_dataset_output["OVERALL_FEATURE_LIST"]
        self.DETAILED_FEATURE_LIST = load_dataset_output["DETAILED_FEATURE_LIST"]

    def economic_distillation(self):
        model = torch.load(self.logdir + "/best_model_" + self.config["DATASET_NAME"] + ".pt").to(self.config["DEVICE"])
        model.eval()

        if self.config["INTERPRETABLE_ANALYSIS_CONFIG"]["Y_NAME"] == "PORTFOLIOS":
            model = ModifiedModel(model).to(self.config["DEVICE"])
            model.eval()

        coef_list = []
        intercept_list = []
        indices_list = []

        if self.config["INTERPRETABLE_ANALYSIS_CONFIG"]["INCLUDE_INTERPRETABILITY_ANALYSIS"]:
            ig = Saliency(model)
            feature_att = torch.zeros(self.NUM_TEST_PERIODS, self.NUM_ASSETS, self.NUM_FEATURES_ORIGINAL)
            every_day_att = torch.zeros(self.NUM_TEST_PERIODS, self.NUM_FEATURES_ORIGINAL)

        for day, data in enumerate(self.test_loader):
            x_data, label = data

            if self.config["INTERPRETABLE_ANALYSIS_CONFIG"]["INCLUDE_INTERPRETABILITY_ANALYSIS"]:
                # Calculate the feature attributions for each asset using saliency method
                for i in range(self.NUM_ASSETS):  # num_feats
                    attributions = ig.attribute(x_data.float(), target=i, abs=False)
                    attributions = attributions.view(1, self.NUM_ASSETS, self.WINDOW_SIZE, self.NUM_FEATURES_ORIGINAL)
                    # attributions.shape = state.shape = torch.Size([1, NUM_ASSETS, WINDOW_SIZE, NUM_FEATURES_ORIGINAL])
                    attributions = attributions[:, i, :, :]  # attributions_sum.shape: torch.Size([1, WINDOW_SIZE, NUM_FEATURES_ORIGINAL])
                    attributions_mean = torch.mean(attributions, dim=(0, 1))  # [1, WINDOW_SIZE, NUM_FEATURES_ORIGINAL] -> [NUM_FEATURES_ORIGINAL]
                    # print(attributions_sum.shape)
                    feature_att[day, i, :] = attributions_mean
                every_day_att[day, :] = torch.mean(feature_att[day, :, :], dim=0)  # [NUM_ASSETS, NUM_FEATURES_ORIGINAL] -> [NUM_FEATURES_ORIGINAL]

            if self.config["INTERPRETABLE_ANALYSIS_CONFIG"]["INCLUDE_ECONOMIC_DISTILLATION"]:
                final_scores = model(x_data.float())

                if self.config["INTERPRETABLE_ANALYSIS_CONFIG"]["PROP_DISTILLED_FEATURES"] != 1:
                    _, indices = torch.topk(every_day_att[day, :], k=int(self.config["INTERPRETABLE_ANALYSIS_CONFIG"]["PROP_DISTILLED_FEATURES"] * self.NUM_FEATURES_ORIGINAL))  # 获取top k索引
                else:
                    indices = torch.arange(0, self.NUM_FEATURES_ORIGINAL)

                flat_x_data = x_data.view(1, self.NUM_ASSETS, self.WINDOW_SIZE, self.NUM_FEATURES_ORIGINAL).squeeze(0)
                linear_x_data = torch.mean(flat_x_data[:, :, indices], dim=1)  # [1, NUM_ASSETS, WINDOW_SIZE, NUM_FEATURES_ORIGINAL] -> [NUM_ASSETS, NUM_FEATURES_ORIGINAL] -> np

                linear_x = linear_x_data.cpu().detach().numpy()
                linear_y = final_scores.squeeze(0).cpu().detach().numpy()

                linear_model = DistillerSelector().select_distiller()
                linear_model.fit(linear_x, linear_y)

                coef = np.zeros(self.NUM_FEATURES_ORIGINAL)  # Initializes a zero vector whose length is the number of features
                real_coef = linear_model.coef_  # Take the actual non-zero coefficient
                coef[indices] = real_coef  # Insert the actual nonzero coefficient into the corresponding position of the zero vector

                # Save to generate a linear portfolio when you iterate over the test set a second time
                coef_list.append(coef)
                intercept_list.append(linear_model.intercept_)
                indices_list.append(indices)

        if self.config["INTERPRETABLE_ANALYSIS_CONFIG"]["INCLUDE_INTERPRETABILITY_ANALYSIS"]:
            # Mean the every_day_att across days
            every_day_att_mean = torch.mean(every_day_att, dim=0)
            # print(every_day_att_mean)

            # Plot the feature attributions using matplotlib.pyplot
            fig = plt.figure(figsize=(9, 4))
            # Plot the average feature attribution for each feature across all time steps as a bar chart
            pre_num_features = 0
            for i, unit in enumerate(self.OVERALL_FEATURE_LIST):
                current_num_features = self.DETAILED_NUM_FEATURES[unit]
                mean_result = torch.mean(every_day_att_mean[pre_num_features: current_num_features+pre_num_features])
                plt.bar(unit, mean_result)
                pre_num_features = current_num_features

            plt.xlabel("Features")
            plt.ylabel("Importance of Features")
            plt.title(self.config["DATASET_NAME"])
            plt.grid(True)
            plt.tight_layout()
            fig.autofmt_xdate()
            plt.savefig(self.logdir + "/" + self.config["MODEL_NAME"] + "_" + self.config["DATASET_NAME"] + "_INTERPRETABILITY_ANALYSIS.pdf",
                        format="pdf",
                        dpi=300,
                        bbox_inches="tight")
            plt.show()

        if self.config["INTERPRETABLE_ANALYSIS_CONFIG"]["INCLUDE_ECONOMIC_DISTILLATION"]:
            portfolios = torch.zeros((self.NUM_TEST_PERIODS, self.NUM_ASSETS))
            labels = torch.zeros((self.NUM_TEST_PERIODS, self.NUM_ASSETS))
            start_time = time.time()
            for day, data in enumerate(self.test_loader):
                x_data, label = data

                coef = coef_list[day]
                intercept = intercept_list[day]
                indices = indices_list[day]

                flat_x_data = x_data.view(1, self.NUM_ASSETS, self.WINDOW_SIZE, self.NUM_FEATURES_ORIGINAL).squeeze(0)
                linear_x_data = torch.mean(flat_x_data, dim=1)  # [1, NUM_ASSETS, WINDOW_SIZE, NUM_FEATURES_ORIGINAL] -> [NUM_ASSETS, NUM_FEATURES_ORIGINAL]

                x_new = linear_x_data.cpu().detach().numpy()
                y_pred = np.dot(x_new, coef) + intercept

                # pred -> portfolio
                y_pred = torch.from_numpy(y_pred).to(self.config["DEVICE"])
                # print(y_pred)
                if self.config["INTERPRETABLE_ANALYSIS_CONFIG"]["Y_NAME"] == "SCORES":
                    portfolio = actual_portfolio_selection(y_pred.unsqueeze(0))
                elif self.config["INTERPRETABLE_ANALYSIS_CONFIG"]["Y_NAME"] == "PORTFOLIOS":
                    positive_y_pred = torch.clamp(y_pred, min=1e-2)
                    portfolio = positive_y_pred / torch.sum(positive_y_pred)

                labels[day, :] = label
                portfolios[day, :] = portfolio

            runtime = time.time() - start_time
            economic_distiller_caculate_metric_output = MetricCaculator(mode="ed").caculate_metric(portfolios, labels, runtime)
            # print(economic_distiller_caculate_metric_output)

            # Convert coefficients to ndarray
            coef_array = np.array(coef_list)
            pre_num_features = 0
            mean_coef_array = np.zeros((self.NUM_TEST_PERIODS, len(self.OVERALL_FEATURE_LIST)))
            for i, unit in enumerate(self.OVERALL_FEATURE_LIST):
                current_num_features = self.DETAILED_NUM_FEATURES[unit]
                mean_result = np.mean(coef_array[:, pre_num_features: current_num_features + pre_num_features], axis=1)
                mean_coef_array[:, i] = mean_result
                pre_num_features = current_num_features

            # Calculate the statistical characteristics of time
            pd.set_option("display.max_column", None)
            print(pd.DataFrame(mean_coef_array).describe().round(4))

            # Draw a coefficient line chart
            plt.figure()
            for i in range(mean_coef_array.shape[1]):
                plt.plot(mean_coef_array[:, i], color="black", marker=self.config["MARKERS"][i], markevery=self.config["MARKEVERY"][self.config["DATASET_NAME"]], alpha=0.5)
            plt.title(self.config["DATASET_NAME"])
            if self.config["PLOT_CHINESE"]:
                plt.ylabel("特征重要程度")
                plt.xlabel("交易期")
            else:
                plt.ylabel("Coefficients")
                plt.xlabel("Trading Periods")
            plt.legend(self.OVERALL_FEATURE_LIST)
            plt.grid(True)
            plt.show()

            # Draw a coefficient heat map
            plt.figure(figsize=(9, 2.5))
            vmax = max(abs(np.max(mean_coef_array)), abs(np.min(mean_coef_array)))
            plt.imshow(mean_coef_array.T, aspect="auto", alpha=0.6, cmap="bwr", vmax=vmax, vmin=-vmax)
            plt.title(self.config["DATASET_NAME"])
            if self.config["PLOT_CHINESE"]:
                plt.ylabel("特征")
                plt.xlabel("交易期")
            else:
                plt.ylabel("Feature")
                plt.xlabel("Trading Periods")

            OVERALL_FEATURE_LIST = []
            if self.config["PLOT_CHINESE"]:
                if self.config["FEATURE_ENGINEERING_CONFIG"]["INCLUDE_OHLCV_FEATURES"]:
                    OVERALL_FEATURE_LIST.append("基础特征")
                if self.config["FEATURE_ENGINEERING_CONFIG"]["INCLUDE_OVERLAP_FEATURES"]:
                    OVERALL_FEATURE_LIST.append("重叠特征")
                if self.config["FEATURE_ENGINEERING_CONFIG"]["INCLUDE_MOMENTUM_FEATURES"]:
                    OVERALL_FEATURE_LIST.append("动量特征")
                if self.config["FEATURE_ENGINEERING_CONFIG"]["INCLUDE_VOLUME_FEATURES"]:
                    OVERALL_FEATURE_LIST.append("量价特征")
                if self.config["FEATURE_ENGINEERING_CONFIG"]["INCLUDE_CYCLE_FEATURES"]:
                    OVERALL_FEATURE_LIST.append("周期特征")
                if self.config["FEATURE_ENGINEERING_CONFIG"]["INCLUDE_PRICE_FEATURES"]:
                    OVERALL_FEATURE_LIST.append("价格特征")
                if self.config["FEATURE_ENGINEERING_CONFIG"]["INCLUDE_VOLATILITY_FEATURES"]:
                    OVERALL_FEATURE_LIST.append("波动特征")
                if self.config["FEATURE_ENGINEERING_CONFIG"]["INCLUDE_PATTERN_FEATURES"]:
                    OVERALL_FEATURE_LIST.append("模式特征")
            else:
                if self.config["FEATURE_ENGINEERING_CONFIG"]["INCLUDE_OHLCV_FEATURES"]:
                    OVERALL_FEATURE_LIST.append("OHLCV Features")
                if self.config["FEATURE_ENGINEERING_CONFIG"]["INCLUDE_OVERLAP_FEATURES"]:
                    OVERALL_FEATURE_LIST.append("Overlap Features")
                if self.config["FEATURE_ENGINEERING_CONFIG"]["INCLUDE_MOMENTUM_FEATURES"]:
                    OVERALL_FEATURE_LIST.append("Momentum Features")
                if self.config["FEATURE_ENGINEERING_CONFIG"]["INCLUDE_VOLUME_FEATURES"]:
                    OVERALL_FEATURE_LIST.append("Volume Features")
                if self.config["FEATURE_ENGINEERING_CONFIG"]["INCLUDE_CYCLE_FEATURES"]:
                    OVERALL_FEATURE_LIST.append("Cycle Features")
                if self.config["FEATURE_ENGINEERING_CONFIG"]["INCLUDE_PRICE_FEATURES"]:
                    OVERALL_FEATURE_LIST.append("Price Features")
                if self.config["FEATURE_ENGINEERING_CONFIG"]["INCLUDE_VOLATILITY_FEATURES"]:
                    OVERALL_FEATURE_LIST.append("Volatility Features")
                if self.config["FEATURE_ENGINEERING_CONFIG"]["INCLUDE_PATTERN_FEATURES"]:
                    OVERALL_FEATURE_LIST.append("Pattern Features")

            plt.yticks(range(len(OVERALL_FEATURE_LIST)), OVERALL_FEATURE_LIST)
            plt.colorbar()
            plt.savefig(self.logdir + "/" + self.config["MODEL_NAME"] + "_" + self.config["DATASET_NAME"] + "_HEATMAP.pdf",
                        format="pdf",
                        dpi=300,
                        bbox_inches="tight")
            plt.show()
            return economic_distiller_caculate_metric_output
        return None


import time
import torch

import matplotlib
import matplotlib.pyplot as plt
import numpy as np

from IPython import display
from tqdm import tqdm
from finol.model_layer.ModelSelector import ModelSelector
from finol.optimization_layer.CriterionSelector import CriterionSelector
from finol.optimization_layer.OptimizerSelector import OptimizerSelector
from finol.optimization_layer.OptunaOptimizer import OptunaOptimizer
from finol.utils import load_config, portfolio_selection, set_seed


class ModelTrainer:
    def __init__(self, load_dataset_output):
        self.config = load_config()
        self.load_dataset_output = load_dataset_output
        self.logdir = load_dataset_output["logdir"]
        self.train_loader = load_dataset_output["train_loader"]  # train_loader \ test_loader_for_train
        self.val_loader = load_dataset_output["val_loader"]  # val_loader \ test_loader_for_train

        self.is_ipython = "inline" in matplotlib.get_backend()
        print(self.is_ipython)

    def plot_loss_notebook(self):
        if self.is_ipython:
            plt.ion()
            plt.figure()  # figsize=(12, 5)
            plt.clf()
            plt.plot(self.avg_train_loss_list, linestyle="-", marker=self.config["MARKERS"][0], markevery=int(self.config["NUM_EPOCHES"]/20), color="black", alpha=0.5, label="train loss")
            plt.plot(self.avg_val_loss_list, linestyle=":", marker=self.config["MARKERS"][1], markevery=int(self.config["NUM_EPOCHES"]/20), color="black", alpha=0.5, label="val loss")
            plt.xlabel("Epoch")
            plt.ylabel("Loss")
            plt.legend()
            plt.grid(True)
            display.clear_output(wait=True)
            display.display(plt.gcf())
            plt.tight_layout()
            # plt.yscale("log")
            plt.savefig(self.logdir + "/" + self.config["DATASET_NAME"] + "_LOSS.pdf",
                        format="pdf",
                        dpi=300,
                        bbox_inches="tight")
            plt.clf()
            plt.close()

    def plot_loss(self):
        # not_ipython = "inline" not in matplotlib.get_backend()
        if not self.is_ipython:
            plt.figure()  # figsize=(12, 5)
            plt.plot(np.array(self.avg_train_loss_list), linestyle="-", marker=self.config["MARKERS"][0], markevery=int(self.config["NUM_EPOCHES"]/20), color="black", alpha=0.5, label="train loss")
            plt.plot(np.array(self.avg_val_loss_list), linestyle=":", marker=self.config["MARKERS"][1], markevery=int(self.config["NUM_EPOCHES"]/20), color="black", alpha=0.5, label="val loss")
            plt.xlabel("Epoch")
            plt.ylabel("Loss")
            plt.legend()
            plt.grid(True)
            plt.tight_layout()
            # plt.yscale("log")
            plt.savefig(self.logdir + "/" + self.config["DATASET_NAME"] + "_LOSS.pdf",
                        format="pdf",
                        dpi=300,
                        bbox_inches="tight")
            plt.show()

    def train_model(self):
        if self.config["TUNE_PARAMETERS"]:
            # model, best_model = OptunaOptimizer(load_dataset_output=self.load_dataset_output).train_via_optuna()
            OptunaOptimizer(load_dataset_output=self.load_dataset_output).train_via_optuna()
            pass

        set_seed(seed=self.config["MANUAL_SEED"])
        model = ModelSelector(self.load_dataset_output).select_model()
        optimizer = OptimizerSelector(model).select_optimizer()
        criterion = CriterionSelector()

        self.train_loss_list = []
        self.val_loss_list = []
        self.avg_train_loss_list = []
        self.avg_val_loss_list = []
        best_val_loss = float("inf")

        for e in tqdm(range(self.config["NUM_EPOCHES"]), desc="Training"):
            model.train()
            train_loss = 0
            for i, data in enumerate(self.train_loader, 1):
                x_data, label = data
                final_scores = model(x_data.float())
                portfolio = portfolio_selection(final_scores)
                loss = criterion(portfolio, label.float())

                optimizer.zero_grad()
                loss.backward()
                optimizer.step()
                train_loss += loss.item()

            train_loss /= len(self.train_loader)
            self.train_loss_list.append(train_loss)
            self.avg_train_loss_list.append(sum(self.train_loss_list) / len(self.train_loss_list))

            if (e + 1) % self.config["SAVE_EVERY"] == 0:
                with torch.no_grad():
                    model.eval()
                    val_loss = 0
                    for i, data in enumerate(self.val_loader, 1):
                        val_data, label = data
                        final_scores = model(val_data.float())
                        portfolio = portfolio_selection(final_scores)
                        loss = criterion(portfolio, label.float())
                        val_loss += loss.item()

                    val_loss /= len(self.val_loader)
                    self.val_loss_list.append(val_loss)
                    self.avg_val_loss_list.append(sum(self.val_loss_list) / len(self.val_loss_list))

                    if val_loss < best_val_loss:
                        best_val_loss = val_loss
                        # best_model = model
                        # print("best_model", "e:", e)
                        torch.save(model, self.logdir + "/best_model_" + self.config["DATASET_NAME"] + ".pt")

            if self.config["PLOT_DYNAMIC_LOSS"]:
                if (e + 1) % 10 == 0:
                    # print("Epoch: {}, Train Loss: {}, Val Loss: {}".format(e + 1, train_loss, val_loss))
                    self.plot_loss_notebook()

        self.plot_loss_notebook()
        self.plot_loss()

        train_model_output = {
            "logdir": self.logdir,
            # "last_model": model,
            # "best_model": best_model
        }
        # print(
        #     best_model
        # )
        # torch.save(best_model, self.logdir + "/best_model_" + self.config["DATASET_NAME"] + ".pt")
        torch.save(model, self.logdir + "/last_model_" + self.config["DATASET_NAME"] + ".pt")
        return train_model_output


if __name__ == "__main__":
    from finol.data_layer.DatasetLoader import DatasetLoader
    from finol.evaluation_layer.ModelEvaluator import ModelEvaluator
    load_dataset_output = DatasetLoader().load_dataset()
    train_model_output = ModelTrainer(load_dataset_output).train_model()
    evaluate_model_output = ModelEvaluator(load_dataset_output, train_model_output).evaluate_model()


from finol.data_layer.DatasetLoader import DatasetLoader
from finol.optimization_layer.ModelTrainer import ModelTrainer
from finol.evaluation_layer.ModelEvaluator import ModelEvaluator


load_dataset_output = DatasetLoader().load_dataset()
train_model_output = ModelTrainer(load_dataset_output).train_model()
evaluate_model_output = ModelEvaluator(load_dataset_output, train_model_output).evaluate_model()


import torch
import torch.nn.functional as F

from functools import wraps
from torch import nn, einsum
from einops import rearrange, repeat
from finol.data_layer.ScalerSelector import ScalerSelector
from finol.utils import load_config

"""
                       Table 7: Hyper-parameters of the LSRE-CAAN framework.
+---------------------------+--------+------------------------------------------------------------------+
| Hyper-parameter           | Choice | Description                                                      |
+---------------------------+--------+------------------------------------------------------------------+
| Depth of net (L)          | 1      | The number of process layers in LSRE.                            |
| Number of latents (M)     | 1      | The number of latents.                                           |
| Latent dimension (D)      | 32     | The size of the latent space.                                    | 
| Number of cross-heads     | 1      | The number of heads for cross-attention.                         | 
| Number of latent-heads    | 1      | The number of heads for latent self-attention.                   | 
| Cross-attention dimension | 64     | The number of dimensions per cross-attention head.               | 
| Self-attention dimension  | 32     | The number of dimensions per latent self-attention head.         | 
| Dropout ratio             | None   | No dropout is used following Jaegle et al. (2022).               | 
| Embedding dimension       | None   | No Embedding layer is used, as illustrated in Section 4.1.       | 
+---------------------------+--------+------------------------------------------------------------------+
| Optimizer                 | LAMB   | An optimizer specifically designed for Transformer-based models. |
| Learning rate             | 0.001  | Parameter of the LAMB optimizer.                                 |
| Weight decay rate         | 0.01   | Parameter of the LAMB optimizer.                                 |
| Training steps            | 10^4   | Training times.                                                  |
| Episode length (T)        | 50     | The length of an episode.                                        |
+---------------------------+--------+------------------------------------------------------------------+
| G                         | m/2    | Half of the assets are identified as winners.                    |
| W                         | 100    | The look-back window size.                                       |
+---------------------------+--------+------------------------------------------------------------------+
"""


def exists(val):
    return val is not None


def default(val, d):
    return val if exists(val) else d


def cache_fn(f):
    cache = None

    @wraps(f)
    def cached_fn(*args, _cache=True, **kwargs):
        if not _cache:
            return f(*args, **kwargs)
        nonlocal cache
        if cache is not None:
            return cache
        cache = f(*args, **kwargs)
        return cache

    return cached_fn


class PreNorm(nn.Module):
    def __init__(self, dim, fn, context_dim=None):
        super().__init__()
        self.fn = fn
        self.norm = nn.LayerNorm(dim)
        self.norm_context = nn.LayerNorm(context_dim) if exists(context_dim) else None

    def forward(self, x, **kwargs):
        x = self.norm(x)

        if exists(self.norm_context):
            context = kwargs["context"]
            normed_context = self.norm_context(context)
            kwargs.update(context=normed_context)

        return self.fn(x, **kwargs)


class QuickGELU(nn.Module):
    def __init__(self):
        super().__init__()

    def forward(self, x):
        return x * torch.sigmoid(1.702 * x)


class FeedForward(nn.Module):
    def __init__(self, dim):
        super().__init__()
        self.net = nn.Sequential(
            nn.Linear(dim, dim),
            # QuickGELU(),
            # nn.Linear(dim, dim)
        )

    def forward(self, x):
        return self.net(x)


class Attention(nn.Module):
    def __init__(self, query_dim, context_dim=None, heads=8, dim_head=64):
        super().__init__()
        inner_dim = dim_head * heads
        context_dim = default(context_dim, query_dim)
        self.scale = dim_head ** -0.5
        self.heads = heads

        self.to_q = nn.Linear(query_dim, inner_dim, bias=False)
        self.to_kv = nn.Linear(context_dim, inner_dim * 2, bias=False)
        self.to_out = nn.Linear(inner_dim, query_dim)  # query_dim = latent_dim

    def forward(self, x, context=None, mask=None):
        h = self.heads

        q = self.to_q(x)
        context = default(context, x)  # return context if exists(context) else x
        k, v = self.to_kv(context).chunk(2, dim=-1)

        q, k, v = map(lambda t: rearrange(t, "b n (h d) -> (b h) n d", h=h), (q, k, v))

        sim = einsum("b i d, b j d -> b i j", q, k) * self.scale

        if exists(mask):
            mask = rearrange(mask, "b ... -> b (...)")
            max_neg_value = -torch.finfo(sim.dtype).max
            mask = repeat(mask, "b j -> (b h) () j", h=h)
            sim.masked_fill_(~mask, max_neg_value)

        # attention, what we cannot get enough of
        attn = sim.softmax(dim=-1)

        out = einsum("b i j, b j d -> b i d", attn, v)
        out = rearrange(out, "(b h) n d -> b n (h d)", h=h)
        return self.to_out(out)


class LSRE(nn.Module):
    r"""
    This class implements the Long Sequence Representations Extractor (LSRE) module

    For more details, please refer to the papers `Online portfolio management via deep reinforcement learning with
    high-frequency data <https://www.sciencedirect.com/science/article/abs/pii/S030645732200348X>` and `Perceiver IO: A
    General Architecture for Structured Inputs & Outputs <https://arxiv.org/abs/2107.14795>`
    """

    def __init__(self, model_args, model_params):
        super().__init__()
        self.latents = nn.Parameter(torch.randn(model_params["NUM_LATENTS"], model_params["LATENT_DIM"]))
        # self.latents = nn.Parameter(torch.zeros(model_params["NUM_LATENTS"], model_params["LATENT_DIM"]))
        self.cross_attend_blocks = nn.ModuleList([
            PreNorm(model_params["LATENT_DIM"], Attention(model_params["LATENT_DIM"], model_args["NUM_FEATURES_ORIGINAL"], heads=model_params["CROSS_HEADS"], dim_head=model_params["CROSS_DIM_HEAD"]), context_dim=model_args["NUM_FEATURES_ORIGINAL"]),
            PreNorm(model_params["LATENT_DIM"], FeedForward(model_params["LATENT_DIM"]))
        ])
        get_latent_attn = lambda: PreNorm(model_params["LATENT_DIM"], Attention(model_params["LATENT_DIM"], heads=model_params["LATENT_HEADS"], dim_head=model_params["LATENT_DIM_HEAD"]))
        get_latent_ff = lambda: PreNorm(model_params["LATENT_DIM"], FeedForward(model_params["LATENT_DIM"]))
        get_latent_attn, get_latent_ff = map(cache_fn, (get_latent_attn, get_latent_ff))

        self.layers = nn.ModuleList([])
        cache_args = {"_cache": True}

        for i in range(model_params["NUM_LAYERS"]):
            self.layers.append(nn.ModuleList([
                get_latent_attn(**cache_args),
                get_latent_ff(**cache_args)
            ]))

    def forward(self, data, mask=None, queries=None):
        # b, *_, device = *data.shape, data.device
        b, *_ = *data.shape, data.device

        # latents
        x = repeat(self.latents, "n d -> b n d", b=b)
        cross_attn, cross_ff = self.cross_attend_blocks

        x = cross_attn(x, context=data, mask=mask) + x
        x = cross_ff(x) + x

        # layers
        for self_attn, self_ff in self.layers:
            x = self_attn(x) + x
            x = self_ff(x) + x

        # return torch.mean(x, dim=1)  # [batch_size * NUM_ASSETS, num_latents, latent_dim] -> [batch_size * NUM_ASSETS, latent_dim]
        return x[:, -1, :].squeeze()  # [batch_size * NUM_ASSETS, num_latents, latent_dim] -> [batch_size * NUM_ASSETS, latent_dim]


class CAAN(nn.Module):
    """
    This class implements the Cross Asset Attention Network (CAAN) module
    """
    def __init__(self, model_params):
        super().__init__()
        self.linear_query = torch.nn.Linear(model_params["LATENT_DIM"], model_params["LATENT_DIM"])
        self.linear_key = torch.nn.Linear(model_params["LATENT_DIM"], model_params["LATENT_DIM"])
        self.linear_value = torch.nn.Linear(model_params["LATENT_DIM"], model_params["LATENT_DIM"])
        self.linear_winner = torch.nn.Linear(model_params["LATENT_DIM"], 1)

    def forward(self, x):
        query = self.linear_query(x)  # [batch_size, num_assets, LATENT_DIM]
        key = self.linear_key(x)  # [batch_size, num_assets, LATENT_DIM]
        value = self.linear_value(x)  # [batch_size, num_assets, LATENT_DIM]

        beta = torch.matmul(query, key.transpose(1, 2)) / torch.sqrt(torch.tensor(float(query.shape[-1])))  # [batch_size, num_assets, LATENT_DIM]
        beta = F.softmax(beta, dim=-1).unsqueeze(-1)
        x = torch.sum(value.unsqueeze(1) * beta, dim=2)  # [batch_size, num_assets, LATENT_DIM]

        final_scores = self.linear_winner(x).squeeze(-1)  # [batch_size, num_assets]
        return final_scores


class LSRE_CAAN(nn.Module):
    r"""
    This class implements the LSRE-CAAN model

    For more details, please refer to the paper `Online portfolio management via deep reinforcement learning with
    high-frequency data <https://www.sciencedirect.com/science/article/abs/pii/S030645732200348X>`
    """
    def __init__(self, model_args, model_params):
        super().__init__()
        self.config = load_config()
        self.model_args = model_args
        self.model_params = model_params

        # self.token_emb = nn.Linear(num_features_original, self.dim)
        self.pos_emb = nn.Embedding(model_args["WINDOW_SIZE"], model_args["NUM_FEATURES_ORIGINAL"])
        self.lsre = LSRE(model_args, model_params)
        self.caan = CAAN(model_params)
        self.dropout = nn.Dropout(p=self.model_params["DROPOUT"])
        if self.config["MODEL_NAME"] == "LSRE-CAAN-d":
            self.ab_study_linear_1 = torch.nn.Linear(num_features_original, self.model_params["LATENT_DIM"])
        if self.config["MODEL_NAME"] == "LSRE-CAAN-dd":
            self.ab_study_linear_2 = torch.nn.Linear(self.model_params["LATENT_DIM"], 1)

    def forward(self, x):
        batch_size, num_assets, num_features_augmented = x.shape
        DEVICE = x.device

        """Input Transformation"""
        x = x.view(batch_size, num_assets, self.model_args["WINDOW_SIZE"], self.model_args["NUM_FEATURES_ORIGINAL"])
        x = rearrange(x, "b m n d -> (b m) n d")
        if self.config["SCALER"].startswith("Window"):
            x = ScalerSelector().window_normalize(x)

        """Long Sequence Representations Extractor (LSRE)"""
        # x = self.token_emb(x)  # optional
        pos_emb = self.pos_emb(torch.arange(self.model_args["WINDOW_SIZE"], device=DEVICE))
        pos_emb = rearrange(pos_emb, "n d -> () n d")
        x = x + pos_emb

        if self.config["MODEL_NAME"] == "LSRE-CAAN-d":
            # stock_rep = torch.mean(x, dim=1)
            stock_rep = rearrange(x, "b n d -> b d n")
            stock_rep = stock_rep[:, :, -1].squeeze(-1)
            stock_rep = self.ab_study_linear_1(stock_rep)
        else:
            stock_rep = self.lsre(x, mask=None, queries=None)  # [batch_size * num_assets, LATENT_DIM]

        stock_rep = self.dropout(stock_rep)
        x = stock_rep.view(batch_size, num_assets, self.model_params["LATENT_DIM"])

        """Cross Asset Attention Network (CAAN)"""
        if self.config["MODEL_NAME"] == "LSRE-CAAN-dd":
            final_scores = self.ab_study_linear_2(x).squeeze(-1)
        else:
            final_scores = self.caan(x)

        return final_scores


if __name__ == "__main__":
    DEVICE = "cuda"
    torch.manual_seed(0)
    batch_size = 2
    num_assets = 2
    window_size = 3
    num_features_original = 4
    num_features_augmented = window_size * num_features_original
    x = torch.rand(batch_size, num_assets, num_features_augmented).to(DEVICE)
    model = LSRE_CAAN(
        num_assets=num_assets,
        num_features_augmented=num_features_augmented,
        num_features_original=num_features_original,
        window_size=window_size,
    ).to(DEVICE)
    final_scores = model(x)
    print(final_scores)



import os
import time
import torch
import pandas as pd

from tqdm import tqdm
from torch.utils.data import TensorDataset
from torch.utils.data import DataLoader
from finol.data_layer.ScalerSelector import ScalerSelector
from finol.utils import ROOT_PATH, load_config, make_logdir, download_data


class DatasetLoader:
    def __init__(self):
        self.config = load_config()

    def data_accessing(self, folder_path):
        raw_files = []
        for file_name in tqdm(os.listdir(folder_path), desc="Data Loading"):
            if file_name.endswith(".xlsx") or file_name.endswith(".xls"):
                file_path = os.path.join(folder_path, file_name)
                try:
                    dataframe = pd.read_excel(file_path)
                    raw_files.append(dataframe)
                except Exception as e:
                    print(f"An error occurred while loading file {file_path}: {str(e)}")
        return raw_files

    def feature_engineering(self, df):
        ohlcv_features_df = pd.DataFrame()
        overlap_features_df = pd.DataFrame()
        momentum_features_df = pd.DataFrame()
        volume_features_df = pd.DataFrame()
        cycle_features_df = pd.DataFrame()
        price_features_df = pd.DataFrame()
        volatility_features_df = pd.DataFrame()
        pattern_features_df = pd.DataFrame()

        if self.config["FEATURE_ENGINEERING_CONFIG"]["INCLUDE_OHLCV_FEATURES"]:
            ohlcv_features = {
                "OPEN": df.OPEN,
                "HIGH": df.HIGH,
                "LOW": df.LOW,
                "CLOSE": df.CLOSE,
                "VOLUME": df.VOLUME
            }
            ohlcv_features_df = pd.DataFrame(ohlcv_features)

        if self.config["FEATURE_ENGINEERING_CONFIG"]["INCLUDE_OVERLAP_FEATURES"]:
            import talib as ta
            overlap_features = {
                "BBANDS_UPPER": ta.BBANDS(df.CLOSE)[0],  # Bollinger Bands - Upper Band
                "BBANDS_MIDDLE": ta.BBANDS(df.CLOSE)[1],  # Bollinger Bands - Middle Band
                "BBANDS_LOWER": ta.BBANDS(df.CLOSE)[2],  # Bollinger Bands - Lower Band
                "DEMA": ta.DEMA(df.CLOSE),  # Double Exponential Moving Average
                "EMA": ta.EMA(df.CLOSE),  # Exponential Moving Average
                "HT_TRENDLINE": ta.HT_TRENDLINE(df.CLOSE),  # Hilbert Transform - Instantaneous Trendline
                "KAMA": ta.KAMA(df.CLOSE),  # Kaufman Adaptive Moving Average
                "MA": ta.MA(df.CLOSE),  # Moving Average
                "MAMA": ta.MAMA(df.CLOSE)[0],  # MESA Adaptive Moving Average - MAMA
                "MAMA_FAMA": ta.MAMA(df.CLOSE)[1],  # MESA Adaptive Moving Average - FAMA
                "MAVP": ta.MAVP(df.CLOSE, df.DATE),  # Moving Average with Variable Period
                "MIDPOINT": ta.MIDPOINT(df.CLOSE),  # MidPoint over Period
                "MIDPRICE": ta.MIDPRICE(df.HIGH, df.LOW),  # Midpoint Price over Period
                "SAR": ta.SAR(df.HIGH, df.LOW),  # Parabolic SAR
                "SAREXT": ta.SAREXT(df.HIGH, df.LOW),  # Parabolic SAR - Extended
                "SMA": ta.SMA(df.CLOSE),  # Simple Moving Average
                "T3": ta.T3(df.CLOSE),  # Triple Exponential Moving Average (T3)
                "TEMA": ta.TEMA(df.CLOSE),  # Triple Exponential Moving Average
                "TRIMA": ta.TRIMA(df.CLOSE),  # Triangular Moving Average
                "WMA": ta.WMA(df.CLOSE)  # Weighted Moving Average
            }
            overlap_features_df = pd.DataFrame(overlap_features)

        if self.config["FEATURE_ENGINEERING_CONFIG"]["INCLUDE_MOMENTUM_FEATURES"]:
            import talib as ta
            momentum_features = {
                "ADX": ta.ADX(df.HIGH, df.LOW, df.CLOSE),  # Average Directional Movement Index
                "ADXR": ta.ADXR(df.HIGH, df.LOW, df.CLOSE),  # Average Directional Movement Index Rating
                "APO": ta.APO(df.CLOSE),  # Absolute Price Oscillator
                "AROON_UP": ta.AROON(df.HIGH, df.LOW)[0],  # Aroon Up
                "AROON_DOWN": ta.AROON(df.HIGH, df.LOW)[1],  # Aroon Down
                "AROONOSC": ta.AROONOSC(df.HIGH, df.LOW),  # Aroon Oscillator
                "BOP": ta.BOP(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Balance Of Power
                "CCI": ta.CCI(df.HIGH, df.LOW, df.CLOSE),  # Commodity Channel Index
                "CMO": ta.CMO(df.CLOSE),  # Chande Momentum Oscillator
                "DX": ta.DX(df.HIGH, df.LOW, df.CLOSE),  # Directional Movement Index
                "MACD": ta.MACD(df.CLOSE)[0],  # Moving Average Convergence/Divergence
                "MACD_SIGNAL": ta.MACD(df.CLOSE)[1],  # MACD Signal Line
                "MACD_HIST": ta.MACD(df.CLOSE)[2],  # MACD Histogram
                "MACDEXT": ta.MACDEXT(df.CLOSE)[0],  # MACD with controllable MA type
                "MACDEXT_SIGNAL": ta.MACDEXT(df.CLOSE)[1],  # MACDEXT Signal Line
                "MACDEXT_HIST": ta.MACDEXT(df.CLOSE)[2],  # MACDEXT Histogram
                "MACDFIX": ta.MACDFIX(df.CLOSE)[0],  # Moving Average Convergence/Divergence Fix 12/26
                "MACDFIX_SIGNAL": ta.MACDFIX(df.CLOSE)[1],  # MACDFIX Signal Line
                "MACDFIX_HIST": ta.MACDFIX(df.CLOSE)[2],  # MACDFIX Histogram
                "MFI": ta.MFI(df.HIGH, df.LOW, df.CLOSE, df.VOLUME),  # Money Flow Index
                "MINUS_DI": ta.MINUS_DI(df.HIGH, df.LOW, df.CLOSE),  # Minus Directional Indicator
                "MINUS_DM": ta.MINUS_DM(df.HIGH, df.LOW),  # Minus Directional Movement
                "MOM": ta.MOM(df.CLOSE),  # Momentum
                "PLUS_DI": ta.PLUS_DI(df.HIGH, df.LOW, df.CLOSE),  # Plus Directional Indicator
                "PLUS_DM": ta.PLUS_DM(df.HIGH, df.LOW),  # Plus Directional Movement
                "PPO": ta.PPO(df.CLOSE),  # Percentage Price Oscillator
                "ROC": ta.ROC(df.CLOSE),  # Rate of change: ((price/prevPrice)-1)*100
                "ROCP": ta.ROCP(df.CLOSE),  # Rate of change Percentage: (price-prevPrice)/prevPrice
                "ROCR": ta.ROCR(df.CLOSE),  # Rate of change ratio: (price/prevPrice)
                "ROCR100": ta.ROCR100(df.CLOSE),  # Rate of change ratio 100 scale: (price/prevPrice)*100
                "RSI": ta.RSI(df.CLOSE),  # Relative Strength Index
                "STOCH_K": ta.STOCH(df.HIGH, df.LOW, df.CLOSE)[0],  # Stochastic %K
                "STOCH_D": ta.STOCH(df.HIGH, df.LOW, df.CLOSE)[1],  # Stochastic %D
                "STOCHF_K": ta.STOCHF(df.HIGH, df.LOW, df.CLOSE)[0],  # Stochastic Fast %K
                "STOCHF_D": ta.STOCHF(df.HIGH, df.LOW, df.CLOSE)[1],  # Stochastic Fast %D
                "STOCHRSI_K": ta.STOCHRSI(df.CLOSE)[0],  # Stochastic RSI %K
                "STOCHRSI_D": ta.STOCHRSI(df.CLOSE)[1],  # Stochastic RSI %D
                "TRIX": ta.TRIX(df.CLOSE),  # 1-day Rate-Of-Change (ROC) of a Triple Smooth EMA
                "ULTOSC": ta.ULTOSC(df.HIGH, df.LOW, df.CLOSE),  # Ultimate Oscillator
                "WILLR": ta.WILLR(df.HIGH, df.LOW, df.CLOSE)  # Williams" %R
            }
            momentum_features_df = pd.DataFrame(momentum_features)

        if self.config["FEATURE_ENGINEERING_CONFIG"]["INCLUDE_VOLUME_FEATURES"]:
            import talib as ta
            volume_features = {
                "AD": ta.AD(df.HIGH, df.LOW, df.CLOSE, df.VOLUME),  # Chaikin A/D Line
                "ADOSC": ta.ADOSC(df.HIGH, df.LOW, df.CLOSE, df.VOLUME),  # Chaikin A/D Oscillator
                "OBV": ta.OBV(df.CLOSE, df.VOLUME)  # On Balance Volume
            }
            volume_features_df = pd.DataFrame(volume_features)

        if self.config["FEATURE_ENGINEERING_CONFIG"]["INCLUDE_CYCLE_FEATURES"]:
            import talib as ta
            cycle_features = {
                "HT_DCPERIOD": ta.HT_DCPERIOD(df.CLOSE),  # Hilbert Transform - Dominant Cycle Period
                "HT_DCPHASE": ta.HT_DCPHASE(df.CLOSE),  # Hilbert Transform - Dominant Cycle Phase
                "HT_PHASOR_INPHASE": ta.HT_PHASOR(df.CLOSE)[0],  # Hilbert Transform - Phasor Components
                "HT_PHASOR_QUADRATURE": ta.HT_PHASOR(df.CLOSE)[1],  # Hilbert Transform - Phasor Components
                "HT_SINE_LEADSINE": ta.HT_SINE(df.CLOSE)[0],  # Hilbert Transform - SineWave
                "HT_SINE_SINEWAVE": ta.HT_SINE(df.CLOSE)[1],  # Hilbert Transform - SineWave
                "HT_TRENDMODE": ta.HT_TRENDMODE(df.CLOSE)  # Hilbert Transform - Trend vs Cycle Mode
            }
            cycle_features_df = pd.DataFrame(cycle_features)

        if self.config["FEATURE_ENGINEERING_CONFIG"]["INCLUDE_PRICE_FEATURES"]:
            import talib as ta
            price_features = {
                "AVGPRICE": ta.AVGPRICE(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Average Price
                "MEDPRICE": ta.MEDPRICE(df.HIGH, df.LOW),  # Median Price
                "TYPPRICE": ta.TYPPRICE(df.HIGH, df.LOW, df.CLOSE),  # Typical Price
                "WCLPRICE": ta.WCLPRICE(df.HIGH, df.LOW, df.CLOSE)  # Weighted Close Price
            }
            price_features_df = pd.DataFrame(price_features)

        if self.config["FEATURE_ENGINEERING_CONFIG"]["INCLUDE_VOLATILITY_FEATURES"]:
            import talib as ta
            volatility_features = {
                "ATR": ta.ATR(df.HIGH, df.LOW, df.CLOSE),  # Average True Range
                "NATR": ta.NATR(df.HIGH, df.LOW, df.CLOSE),  # Normalized Average True Range
                "TRANGE": ta.TRANGE(df.HIGH, df.LOW, df.CLOSE)  # True Range
            }
            volatility_features_df = pd.DataFrame(volatility_features)

        if self.config["FEATURE_ENGINEERING_CONFIG"]["INCLUDE_PATTERN_FEATURES"]:
            import talib as ta
            pattern_features = {
                "CDL2CROWS": ta.CDL2CROWS(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Two Crows
                "CDL3BLACKCROWS": ta.CDL3BLACKCROWS(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Three Black Crows
                "CDL3INSIDE": ta.CDL3INSIDE(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Three Inside Up/Down
                "CDL3LINESTRIKE": ta.CDL3LINESTRIKE(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Three-Line Strike
                "CDL3OUTSIDE": ta.CDL3OUTSIDE(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Three Outside Up/Down
                "CDL3STARSINSOUTH": ta.CDL3STARSINSOUTH(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Three Stars In The South
                "CDL3WHITESOLDIERS": ta.CDL3WHITESOLDIERS(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Three Advancing White Soldiers
                "CDLABANDONEDBABY": ta.CDLABANDONEDBABY(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Abandoned Baby
                "CDLADVANCEBLOCK": ta.CDLADVANCEBLOCK(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Advance Block
                "CDLBELTHOLD": ta.CDLBELTHOLD(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Belt-Hold
                "CDLBREAKAWAY": ta.CDLBREAKAWAY(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Breakaway
                "CDLCLOSINGMARUBOZU": ta.CDLCLOSINGMARUBOZU(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Closing Marubozu
                "CDLCONCEALBABYSWALL": ta.CDLCONCEALBABYSWALL(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Concealing Baby Swallow
                "CDLCOUNTERATTACK": ta.CDLCOUNTERATTACK(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Counterattack
                "CDLDARKCLOUDCOVER": ta.CDLDARKCLOUDCOVER(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Dark Cloud Cover
                "CDLDOJI": ta.CDLDOJI(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Doji,
                "CDLDOJISTAR": ta.CDLDOJISTAR(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Doji Star
                "CDLDRAGONFLYDOJI": ta.CDLDRAGONFLYDOJI(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Dragonfly Doji
                "CDLENGULFING": ta.CDLENGULFING(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Engulfing Pattern
                "CDLEVENINGDOJISTAR": ta.CDLEVENINGDOJISTAR(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Evening Doji Star
                "CDLEVENINGSTAR": ta.CDLEVENINGSTAR(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Evening Star
                "CDLGAPSIDESIDEWHITE": ta.CDLGAPSIDESIDEWHITE(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Up/Down-Gap Side-By-Side White Lines
                "CDLGRAVESTONEDOJI": ta.CDLGRAVESTONEDOJI(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Gravestone Doji
                "CDLHAMMER": ta.CDLHAMMER(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Hammer
                "CDLHANGINGMAN": ta.CDLHANGINGMAN(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Hanging Man
                "CDLHARAMI": ta.CDLHARAMI(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Harami Pattern
                "CDLHARAMICROSS": ta.CDLHARAMICROSS(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Harami Cross Pattern
                "CDLHIGHWAVE": ta.CDLHIGHWAVE(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # High-Wave Candle
                "CDLHIKKAKE": ta.CDLHIKKAKE(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Hikkake Pattern
                "CDLHIKKAKEMOD": ta.CDLHIKKAKEMOD(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Modified Hikkake Pattern
                "CDLHOMINGPIGEON": ta.CDLHOMINGPIGEON(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Homing Pigeon
                "CDLIDENTICAL3CROWS": ta.CDLIDENTICAL3CROWS(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Identical Three Crows
                "CDLINNECK": ta.CDLINNECK(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # In-Neck Pattern
                "CDLINVERTEDHAMMER": ta.CDLINVERTEDHAMMER(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Inverted Hammer
                "CDLKICKING": ta.CDLKICKING(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Kicking
                "CDLKICKINGBYLENGTH": ta.CDLKICKINGBYLENGTH(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Kicking - Bull/Bear Determined by the Longer Marubozu
                "CDLLADDERBOTTOM": ta.CDLLADDERBOTTOM(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Ladder Bottom
                "CDLLONGLEGGEDDOJI": ta.CDLLONGLEGGEDDOJI(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Long Legged Doji
                "CDLLONGLINE": ta.CDLLONGLINE(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Long Line Candle
                "CDLMARUBOZU": ta.CDLMARUBOZU(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Marubozu
                "CDLMATCHINGLOW": ta.CDLMATCHINGLOW(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Matching Low
                "CDLMATHOLD": ta.CDLMATHOLD(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Mat Hold
                "CDLMORNINGDOJISTAR": ta.CDLMORNINGDOJISTAR(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Morning Doji Star
                "CDLMORNINGSTAR": ta.CDLMORNINGSTAR(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Morning Star
                "CDLONNECK": ta.CDLONNECK(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # On-Neck Pattern
                "CDLPIERCING": ta.CDLPIERCING(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Piercing Pattern
                "CDLRICKSHAWMAN": ta.CDLRICKSHAWMAN(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Rickshaw Man
                "CDLRISEFALL3METHODS": ta.CDLRISEFALL3METHODS(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Rising/Falling Three Methods
                "CDLSEPARATINGLINES": ta.CDLSEPARATINGLINES(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Separating Lines
                "CDLSHOOTINGSTAR": ta.CDLSHOOTINGSTAR(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Shooting Star
                "CDLSHORTLINE": ta.CDLSHORTLINE(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Short Line Candle
                "CDLSPINNINGTOP": ta.CDLSPINNINGTOP(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Spinning Top
                "CDLSTALLEDPATTERN": ta.CDLSTALLEDPATTERN(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Stalled Pattern
                "CDLSTICKSANDWICH": ta.CDLSTICKSANDWICH(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Stick Sandwich
                "CDLTAKURI": ta.CDLTAKURI(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Takuri (Dragonfly Doji with Very Long Lower Shadow)
                "CDLTASUKIGAP": ta.CDLTASUKIGAP(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Tasuki Gap
                "CDLTHRUSTING": ta.CDLTHRUSTING(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Thrusting Pattern
                "CDLTRISTAR": ta.CDLTRISTAR(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Tristar Pattern
                "CDLUNIQUE3RIVER": ta.CDLUNIQUE3RIVER(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Unique 3 River
                "CDLUPSIDEGAP2CROWS": ta.CDLUPSIDEGAP2CROWS(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Upside Gap Two Crows
                "CDLXSIDEGAP3METHODS": ta.CDLXSIDEGAP3METHODS(df.OPEN, df.HIGH, df.LOW, df.CLOSE)  # Upside/Downside Gap Three Methods
            }
            pattern_features_df = pd.DataFrame(pattern_features)

        _ = pd.concat([df["DATE"], ohlcv_features_df, overlap_features_df, momentum_features_df, volume_features_df, cycle_features_df,
                        price_features_df, volatility_features_df, pattern_features_df], axis=1)
        _.set_index("DATE", inplace=True)
        DETAILED_FEATURE_LIST = _.columns.tolist()
        DETAILED_NUM_FEATURES = {
            "OHLCV_FEATURES": ohlcv_features_df.shape[1],
            "OVERLAP_FEATURES": overlap_features_df.shape[1],
            "MOMENTUM_FEATURES": momentum_features_df.shape[1],
            "VOLUME_FEATURES": volume_features_df.shape[1],
            "CYCLE_FEATURES": cycle_features_df.shape[1],
            "PRICE_FEATURES": price_features_df.shape[1],
            "VOLATILITY_FEATURES": volatility_features_df.shape[1],
            "PATTERN_FEATURES": pattern_features_df.shape[1],
        }
        return _, DETAILED_FEATURE_LIST, DETAILED_NUM_FEATURES

    def data_augmentation(self, df):
        if self.config["DATA_AUGMENTATION_CONFIG"]["WINDOW_DATA"]["INCLUDE_WINDOW_DATA"]:
            WINDOW_SIZE = self.config["DATA_AUGMENTATION_CONFIG"]["WINDOW_DATA"]["WINDOW_SIZE"]
            df = pd.concat([df] + [df.shift(i).add_prefix(f"prev_{i}_") for i in range(1, WINDOW_SIZE)], axis=1)
        else:
            WINDOW_SIZE = 1
        return df, WINDOW_SIZE

    def data_cleaning(self, df):
        """
        Clean the DataFrame by removing rows with missing values.
        """
        return df.dropna(how="any")

    def zscore_calculation(self, df):
        df_normalized = df.copy()
        numeric_features = df.select_dtypes(include=["int", "float"]).columns
        scaler = ScalerSelector().select_scaler()
        if scaler != None:
            zscore = scaler.fit(df_normalized[numeric_features])  # zscore is different for different assets
        else:
            zscore = None
        return zscore

    def data_normalization(self, df, zscore):
        """
        Normalize all numeric features in DataFrame
        """
        df_normalized = df.copy()
        numeric_features = df.select_dtypes(include=["int", "float"]).columns
        if zscore != None:
            df_normalized[numeric_features] = zscore.transform(df_normalized[numeric_features])
            # print(df_normalized[numeric_features])
            # print(df_normalized[numeric_features]["OPEN"].values)
            # print(df_normalized[numeric_features]["OPEN"].values.mean())  # -3.620599927440001e-16
            # print(df_normalized[numeric_features]["OPEN"].values.std())  # 1
            # time.sleep(1111)
        else:
            df_normalized[numeric_features] = df_normalized[numeric_features]
        return df_normalized

    def data_splitting(self, df):
        """
        Split the DataFrame into train, validation, and test sets.
        """
        train_start = pd.to_datetime(self.config["DATASET_SPLIT_CONFIG"][self.config["DATASET_NAME"]]["TRAIN_START_TIMESTAMP"])
        train_end = pd.to_datetime(self.config["DATASET_SPLIT_CONFIG"][self.config["DATASET_NAME"]]["TRAIN_END_TIMESTAMP"])
        val_start = pd.to_datetime(self.config["DATASET_SPLIT_CONFIG"][self.config["DATASET_NAME"]]["VAL_START_TIMESTAMP"])
        val_end = pd.to_datetime(self.config["DATASET_SPLIT_CONFIG"][self.config["DATASET_NAME"]]["VAL_END_TIMESTAMP"])
        test_start = pd.to_datetime(self.config["DATASET_SPLIT_CONFIG"][self.config["DATASET_NAME"]]["TEST_START_TIMESTAMP"])
        test_end = pd.to_datetime(self.config["DATASET_SPLIT_CONFIG"][self.config["DATASET_NAME"]]["TEST_END_TIMESTAMP"])

        train = df[df.index <= train_end]
        val = df[(df.index >= val_start) & (df.index <= val_end)]
        test = df[(df.index >= test_start) & (df.index <= test_end)]
        return train, val, test

    def label_making(self, raw_df, df):
        raw_df.set_index("DATE", inplace=True)
        df_label = pd.DataFrame({"LABEL": raw_df.CLOSE.shift(-1) / raw_df.CLOSE})
        df_label.iloc[-1, -1] = 1

        df_label = pd.merge(df, df_label, left_index=True, right_index=True, how="outer")
        df_label = df_label.dropna(how="any")
        df_label = df_label["LABEL"].to_frame()
        return df_label

    def load_dataset(self):
        """
        Load the raw dataset and perform some data processing operations
    
        Returns:
        - df (DataFrame): Processed DataFrame
        """
        # print(self.config["MANUAL_SEED"])
        logdir = make_logdir()
        download_data()

        if self.config["LOAD_LOCAL_DATALOADER"]:
            try:
                load_dataset_output = torch.load(ROOT_PATH + "/data/datasets/" + self.config["DATASET_NAME"] + "_load_dataset_output.pt")
                load_dataset_output["logdir"] = logdir
                print("Local dataloader loaded successfully!")
            except Exception:
                raise ValueError(
                    f"Local dataloader does not exist, please modify LOAD_LOCAL_DATALOADER as `False`"
                )
        else:
            ds_train = []
            ds_val = []
            ds_test = []
            label_train = []
            label_val = []
            label_test = []
            df_label_MATLAB = pd.DataFrame()

            raw_files = self.data_accessing(ROOT_PATH + "/data/datasets/" + self.config["DATASET_NAME"])
            if len(raw_files) > 0:
                # feature_engineering(excel_files)
                for i, df in tqdm(enumerate(raw_files), total=len(raw_files), desc="Data Processing"):
                    raw_df = df.copy()
                    df = self.data_cleaning(df)

                    df, DETAILED_FEATURE_LIST, DETAILED_NUM_FEATURES = self.feature_engineering(df)
                    df = self.data_cleaning(df)

                    df, WINDOW_SIZE = self.data_augmentation(df)
                    df = self.data_cleaning(df)

                    df_label = self.label_making(raw_df, df)

                    # df_label_temp = df_label.copy()
                    # test_start = pd.to_datetime(DATASET_SPLIT_CONFIG.get(dataset_name)["TEST_START_TIMESTAMP"])
                    # test_end = pd.to_datetime(DATASET_SPLIT_CONFIG.get(dataset_name)["TEST_END_TIMESTAMP"])
                    # df_label_temp = df_label_temp[(df_label_temp.index >= test_start) & (df_label_temp.index <= test_end)]
                    # df_label_MATLAB = pd.concat([df_label_MATLAB, df_label_temp], axis=1)

                    train, val, test = self.data_splitting(df)
                    train_label, val_label, test_label = self.data_splitting(df_label)

                    zscore_train = self.zscore_calculation(train)
                    train_normalization = self.data_normalization(train, zscore_train)
                    val_normalization = self.data_normalization(val, zscore_train)
                    test_normalization = self.data_normalization(test, zscore_train)

                    nan_count = df.isna().sum().sum()
                    if nan_count != 0:
                        print(f"nan_count: {nan_count}")
                        time.sleep(1111)

                    nan_count = df.isnull().sum().sum()
                    if nan_count != 0:
                        print(f"nan_count: {nan_count}")
                        time.sleep(1111)

                    train_normalization = self.data_cleaning(train_normalization)
                    val_normalization = self.data_cleaning(val_normalization)
                    test_normalization = self.data_cleaning(test_normalization)

                    ds_train.append(torch.from_numpy(train_normalization.values))
                    ds_val.append(torch.from_numpy(val_normalization.values))
                    ds_test.append(torch.from_numpy(test_normalization.values))
                    label_train.append(torch.from_numpy(train_label["LABEL"].values))
                    label_val.append(torch.from_numpy(val_label["LABEL"].values))
                    label_test.append(torch.from_numpy(test_label["LABEL"].values))

                ds_train = torch.stack(ds_train).permute(1, 0, 2).to(self.config["DEVICE"])  # [num_assets, num_train_periods, num_feats] -> [num_train_periods, num_assets, num_feats]
                ds_val = torch.stack(ds_val).permute(1, 0, 2).to(self.config["DEVICE"])
                ds_test = torch.stack(ds_test).permute(1, 0, 2).to(self.config["DEVICE"])
                label_train = torch.stack(label_train).transpose(0, 1).to(self.config["DEVICE"])  # [num_assets, num_train_periods] -> [num_train_periods, num_assets]
                label_val = torch.stack(label_val).transpose(0, 1).to(self.config["DEVICE"])
                label_test = torch.stack(label_test).transpose(0, 1).to(self.config["DEVICE"])

                train_ids = TensorDataset(ds_train, label_train)
                val_ids = TensorDataset(ds_val, label_val)
                test_ids = TensorDataset(ds_test, label_test)

                # Save the price relative tensor to a mat file with the data set name
                # from scipy import io
                # io.savemat("price_relative_" + dataset_name + ".mat", {"data": df_label_MATLAB.values})
            else:
                print("No Excel files found.")

            load_dataset_output = {
                "logdir": logdir,
                "train_loader": DataLoader(train_ids, batch_size=self.config["BATCH_SIZE"][self.config["DATASET_NAME"]], shuffle=False, drop_last=False),
                "val_loader": DataLoader(val_ids, batch_size=self.config["BATCH_SIZE"][self.config["DATASET_NAME"]], shuffle=False, drop_last=False),
                "test_loader": DataLoader(test_ids, batch_size=1, shuffle=False, drop_last=False),
                "test_loader_for_train": DataLoader(test_ids, batch_size=self.config["BATCH_SIZE"][self.config["DATASET_NAME"]], shuffle=False, drop_last=False),
                "NUM_TRAIN_PERIODS": ds_train.shape[0],
                "NUM_VAL_PERIODS": ds_val.shape[0],
                "NUM_TEST_PERIODS": ds_test.shape[0],
                "NUM_ASSETS": ds_train.shape[1],
                "NUM_FEATURES_AUGMENTED": ds_train.shape[2],
                "NUM_FEATURES_ORIGINAL": int(ds_train.shape[2] / (WINDOW_SIZE)),
                "DETAILED_NUM_FEATURES": DETAILED_NUM_FEATURES,
                "WINDOW_SIZE": WINDOW_SIZE,
                "OVERALL_FEATURE_LIST": [feature[8:] for feature, include in self.config["FEATURE_ENGINEERING_CONFIG"].items() if include],
                "DETAILED_FEATURE_LIST": DETAILED_FEATURE_LIST,

            }
            torch.save(load_dataset_output, ROOT_PATH + "/data/datasets/" + self.config["DATASET_NAME"] + "_load_dataset_output.pt")
        print(load_dataset_output)
        return load_dataset_output


if __name__ == "__main__":
    load_dataset_output = DatasetLoader().load_dataset()
    # print(load_dataset_output)


import pandas as pd
import numpy as np

from matplotlib.patches import Circle, RegularPolygon
from matplotlib.path import Path
from matplotlib.projections import register_projection
from matplotlib.projections.polar import PolarAxes
from matplotlib.spines import Spine
from matplotlib.transforms import Affine2D
from finol.utils import get_variable_name


def radar_factory(num_vars, frame='circle'):
    """
    Create a radar chart with `num_vars` axes.

    This function creates a RadarAxes projection and registers it.

    Parameters
    ----------
    num_vars : int
        Number of variables for radar chart.
    frame : {'circle', 'polygon'}
        Shape of frame surrounding axes.

    """
    # calculate evenly-spaced axis angles
    theta = np.linspace(0, 2 * np.pi, num_vars, endpoint=False)

    class RadarTransform(PolarAxes.PolarTransform):

        def transform_path_non_affine(self, path):
            # Paths with non-unit interpolation steps correspond to gridlines,
            # in which case we force interpolation (to defeat PolarTransform's
            # autoconversion to circular arcs).
            if path._interpolation_steps > 1:
                path = path.interpolated(num_vars)
            return Path(self.transform(path.vertices), path.codes)

    class RadarAxes(PolarAxes):

        name = 'radar'
        PolarTransform = RadarTransform

        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            # rotate plot such that the first axis is at the top
            self.set_theta_zero_location('N')

        def fill(self, *args, closed=True, **kwargs):
            """Override fill so that line is closed by default"""
            return super().fill(closed=closed, *args, **kwargs)

        def plot(self, *args, **kwargs):
            """Override plot so that line is closed by default"""
            lines = super().plot(*args, **kwargs)
            for line in lines:
                self._close_line(line)

        def _close_line(self, line):
            x, y = line.get_data()
            # FIXME: markers at x[0], y[0] get doubled-up
            if x[0] != x[-1]:
                x = np.append(x, x[0])
                y = np.append(y, y[0])
                line.set_data(x, y)

        def set_varlabels(self, labels):
            self.set_thetagrids(np.degrees(theta), labels)

        def _gen_axes_patch(self):
            # The Axes patch must be centered at (0.5, 0.5) and of radius 0.5
            # in axes coordinates.
            if frame == 'circle':
                return Circle((0.5, 0.5), 0.5)
            elif frame == 'polygon':
                return RegularPolygon((0.5, 0.5), num_vars,
                                      radius=.5, edgecolor="k")
            else:
                raise ValueError("Unknown value for 'frame': %s" % frame)

        def _gen_axes_spines(self):
            if frame == 'circle':
                return super()._gen_axes_spines()
            elif frame == 'polygon':
                # spine_type must be 'left'/'right'/'top'/'bottom'/'circle'.
                spine = Spine(axes=self,
                              spine_type='circle',
                              path=Path.unit_regular_polygon(num_vars))
                # unit_regular_polygon gives a polygon of radius 1 centered at
                # (0, 0) but we want a polygon of radius 0.5 centered at (0.5,
                # 0.5) in axes coordinates.
                spine.set_transform(Affine2D().scale(.5).translate(.5, .5)
                                    + self.transAxes)
                return {'polar': spine}
            else:
                raise ValueError("Unknown value for 'frame': %s" % frame)

    register_projection(RadarAxes)
    return theta


def plot_radar_chart(final_profit_result, final_risk_result, column_names, logdir):
    final_profit_result = final_profit_result[["Metric"] + column_names].set_index("Metric")
    final_risk_result = final_risk_result[["Metric"] + column_names].set_index("Metric")
    df = pd.concat([final_profit_result, final_risk_result], ignore_index=False)
    num_columns = len(df.columns)

    data = []
    for strategy in PLOT_ALL_1:
        _ = [df[strategy]["CW"], -df[strategy]["MDD"], -df[strategy]["VR"], df[strategy]["SR"], df[strategy]["APY"]]
        data += [_]

    theta = radar_factory(num_vars=5, frame='polygon')
    labels = ['CW', '- MDD', '- VR', 'SR', 'APY']

    fig, ax = plt.subplots(subplot_kw=dict(projection='radar'))
    colors = ['black'] * (num_columns - 1) + ['red'] * 1

    # plot the four cases from the example data on separate axes
    ax.set_rgrids([0.2, 0.4, 0.6, 0.8])

    # scale the data to [0, 1]
    data_array = np.array(data)
    min_val = np.min(data_array, axis=0)
    max_val = np.max(data_array, axis=0)
    scaled_data = (data_array - min_val) / (max_val - min_val)
    case_data = scaled_data

    for d, color, marker in zip(case_data, colors, MARKERS):
        ax.plot(theta, d, color=color, alpha=0.5, marker=marker)
        ax.fill(theta, d, facecolor=color, alpha=0.15, label='_nolegend_')
    ax.set_varlabels(labels)

    # add legend relative to top-left plot
    plt.title(DATASET_NAME)
    ax.legend(df.columns, loc=(0.7, 0.75))
    plt.tight_layout()
    plt.savefig(logdir + '/' + MODEL_NAME + '_' + DATASET_NAME + '_' + get_variable_name(column_names) + '_RADAR.pdf',
                format='pdf',
                dpi=300,
                bbox_inches='tight')
    plt.show()



import torch.nn as nn

from einops import rearrange
from finol.data_layer.ScalerSelector import ScalerSelector
from finol.utils import load_config


class LSTM(nn.Module):
    def __init__(self, model_args, model_params):
        super().__init__()
        self.config = load_config()
        self.model_args = model_args
        self.model_params = model_params

        self.lstm = nn.LSTM(input_size=model_args["NUM_FEATURES_ORIGINAL"], hidden_size=model_params["HIDDEN_SIZE"],
                            num_layers=model_params["NUM_LAYERS"], batch_first=True)
        self.dropout = nn.Dropout(model_params["DROPOUT"])
        self.fc = nn.Linear(model_params["HIDDEN_SIZE"], 1)

    def forward(self, x):
        batch_size, num_assets, num_features_augmented = x.shape

        """Input Transformation"""
        x = x.view(batch_size, num_assets, self.model_args["WINDOW_SIZE"], self.model_args["NUM_FEATURES_ORIGINAL"])
        x = rearrange(x, "b m n d -> (b m) n d")
        if self.config["SCALER"].startswith("Window"):
            x = ScalerSelector().window_normalize(x)

        """Temporal Representation Extraction"""
        out, _ = self.lstm(x)
        out = self.dropout(out)
        out = out[:, -1, :]

        """Final Scores for Assets"""
        out = out.view(batch_size, num_assets, self.model_params["HIDDEN_SIZE"])
        final_scores = self.fc(out).squeeze(-1)
        return final_scores


import torch

from sklearn import preprocessing
from finol.utils import load_config


def WindowStandardScaler(data):
    """
    Normalize the input tensor data along the window dimension via StandardScaler.

    Input shape:
        data: (..., window_size, feature_dim)

    Output shape:
        standardized_data: (..., window_size, feature_dim)
    """
    window_mean = torch.mean(data, dim=-2, keepdim=True)
    window_std = torch.std(data, dim=-2, keepdim=True) + 1e-7
    standardized_data = (data - window_mean) / window_std
    return standardized_data


def WindowMinMaxScaler(data):
    """
    Normalize the input tensor data along the window dimension via MinMaxScaler.

    Input shape:
        data: (..., window_size, feature_dim)

    Output shape:
        normalized_data: (..., window_size, feature_dim)
    """
    min_vals = torch.min(data, dim=-2, keepdim=True).values
    max_vals = torch.max(data, dim=-2, keepdim=True).values
    normalized_data = (data - min_vals) / (max_vals - min_vals + 1e-7)
    return normalized_data


def WindowMaxAbsScaler(data):
    """
    Normalize the input tensor data along the window dimension via MaxAbsScaler.

    Input shape:
        data: (..., window_size, feature_dim)

    Output shape:
        normalized_data: (..., window_size, feature_dim)
    """
    max_abs_vals = torch.max(torch.abs(data), dim=-2, keepdim=True).values
    normalized_data = data / (max_abs_vals + 1e-7)
    return normalized_data


def WindowRobustScaler(data):
    """
    Normalize the input tensor data along the window dimension via RobustScaler.

    Input shape:
        data: (..., window_size, feature_dim)

    Output shape:
        scaled_data: (..., window_size, feature_dim)
    """

    # Calculate the median and IQR (interquartile range) along the window dimension
    window_median = torch.median(data, dim=-2, keepdim=True).values
    q1 = torch.quantile(data, 0.25, dim=-2, keepdim=True)
    q3 = torch.quantile(data, 0.75, dim=-2, keepdim=True)
    window_iqr = q3 - q1 + 1e-7

    # Scale the data using the median and IQR
    normalized_data = (data - window_median) / window_iqr
    return normalized_data


scaler_dict = {
    "None": "None",
    "StandardScaler": preprocessing.StandardScaler,
    "MinMaxScaler": preprocessing.MinMaxScaler,
    "MaxAbsScaler": preprocessing.MaxAbsScaler,
    "RobustScaler": preprocessing.RobustScaler,
    "WindowStandardScaler": WindowStandardScaler,
    "WindowMinMaxScaler": WindowMinMaxScaler,
    "WindowMaxAbsScaler": WindowMaxAbsScaler,
    "WindowRobustScaler": WindowRobustScaler,
}


class ScalerSelector():
    def __init__(self):
        self.config = load_config()

    def select_scaler(self):
        scaler_cls = scaler_dict.get(self.config["SCALER"], None)
        if scaler_cls is None:
            raise ValueError(f"Invalid scaler name: {self.config['SCALER']}. Supported scalers are: {scaler_dict.keys()}")

        if self.config["SCALER"] == "None" or self.config["SCALER"].startswith("Window"):
            return None

        scaler = scaler_cls()
        return scaler

    def window_normalize(self, data):
        """
        Normalize the input tensor data along the window dimension based on different methods.

        Input shape:
            data: (..., window_size, feature_dim)

        Output shape:
            scaled_data: (..., window_size, feature_dim)
        """
        scaler_cls = scaler_dict.get(self.config["SCALER"], None)
        if scaler_cls is None:
            raise ValueError(f"Invalid scaler name: {self.config['SCALER']}. Supported scalers are: {scaler_dict.keys()}")

        normalized_data = scaler_cls(data)
        return normalized_data





import torch
import optuna
import matplotlib.pyplot as plt

from tqdm import tqdm
from finol.model_layer.ModelSelector import ModelSelector
from finol.optimization_layer.CriterionSelector import CriterionSelector
from finol.optimization_layer.OptimizerSelector import OptimizerSelector
from finol.utils import load_config, update_config, portfolio_selection, set_seed


class OptunaOptimizer:
    def __init__(self, load_dataset_output):
        super().__init__()
        self.config = load_config()
        self.load_dataset_output = load_dataset_output
        self.train_loader = load_dataset_output["train_loader"]
        self.val_loader = load_dataset_output["val_loader"]
        self.logdir = load_dataset_output["logdir"]

    def sample_params(self, trial: optuna.Trial):
        model_sampled_params = self.config["MODEL_PARAMS_SPACE"][self.config["MODEL_NAME"]]
        self.sampled_params = {}

        for param_name, param_space in model_sampled_params.items():
            param_type = param_space["type"]
            param_range = param_space["range"]
            param_step = param_space["step"]

            if param_type == "int":
                self.sampled_params[param_name] = trial.suggest_int(param_name, *param_range, step=param_step)
            elif param_type == "float":
                self.sampled_params[param_name] = trial.suggest_float(param_name, *param_range, step=param_step)
            else:
                raise ValueError(f"Invalid parameter type: {param_type}")

    def objective(self, trial: optuna.Trial):
        self.sample_params(trial)
        set_seed(seed=self.config["MANUAL_SEED"])

        model = ModelSelector(self.load_dataset_output).select_model(self.sampled_params)
        optimizer = OptimizerSelector(model).select_optimizer()
        criterion = CriterionSelector()

        train_loss_list = []
        val_loss_list = []
        best_val_loss = float("inf")

        for e in tqdm(range(self.config["NUM_EPOCHES"]), desc="Training"):
            model.train()
            train_loss = 0
            for i, data in enumerate(self.train_loader, 1):
                x_data, label = data
                final_scores = model(x_data.float())
                portfolio = portfolio_selection(final_scores)

                loss = criterion(portfolio, label.float())

                optimizer.zero_grad()
                loss.backward()
                optimizer.step()
                train_loss += loss.item()

            train_loss /= len(self.train_loader)
            train_loss_list.append(train_loss)

            if (e + 1) % self.config["SAVE_EVERY"] == 0:
                with torch.no_grad():
                    model.eval()
                    val_loss = 0
                    for i, data in enumerate(self.val_loader, 1):
                        val_data, label = data
                        final_scores = model(val_data.float())
                        portfolio = portfolio_selection(final_scores)
                        loss = criterion(portfolio, label.float())
                        val_loss += loss.item()

                    val_loss /= len(self.val_loader)
                    val_loss_list.append(val_loss)
                    # value = sum(val_loss_list) / len(val_loss_list)
                    value = min(val_loss_list)
                    # value = val_loss  # real time val loss

                    if val_loss < best_val_loss:
                        best_val_loss = val_loss

            # Report intermediate objective value.
            trial.report(value, e)

            # Handle pruning based on the intermediate value.
            if trial.should_prune():
                raise optuna.TrialPruned()

        return value

    def train_via_optuna(self):
        # Creating Optuna object and defining its parameters
        self.study = optuna.create_study(
            direction="minimize", sampler=optuna.samplers.TPESampler(seed=self.config["MANUAL_SEED"]),
            pruner=optuna.pruners.MedianPruner(),
            storage="sqlite:///" + self.logdir + "/" + self.config["MODEL_NAME"] + "_" + self.config["DATASET_NAME"] + "_OPTUNA_" + self.config["MODEL_NAME"] + ".db"
        )
        print(f"Sampler is {self.study.sampler.__class__.__name__}")
        self.study.optimize(self.objective, n_trials=self.config["NUM_TRIALS"])

        # To further visualize the results, you can upload the generated {MODEL_NAME}.db file to the Optuna Dashboard:
        # https://optuna.github.io/optuna-dashboard/
        optuna.visualization.plot_intermediate_values(self.study).show()
        optuna.visualization.plot_optimization_history(self.study).show()

        # Showing optimization results
        print("Number of finished trials:", len(self.study.trials))
        print("Best trial parameters:", self.study.best_trial.params)
        print("Best score:", self.study.best_value)

        # Write config
        self.config["MODEL_PARAMS"][self.config["MODEL_NAME"]] = self.study.best_trial.params
        update_config(self.config)


if __name__ == "__main__":
    from finol.data_layer.DatasetLoader import DatasetLoader
    load_dataset_output = DatasetLoader().load_dataset()
    OptunaOptimizer(load_dataset_output=load_dataset_output).train_via_optuna()



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html


from datetime import datetime

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = 'FinOL'
# copyright = f'2024–{datetime.now().year}, FinOL contributors'
copyright = f'2024, FinOL contributors'
author = 'FinOL contributors'
release = 'MIT'

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

# extensions = ["recommonmark", "sphinx_markdown_tables", "sphinx_rtd_theme"]

templates_path = ['_templates']
exclude_patterns = []



# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

import sphinx_rtd_theme

html_theme = "sphinx_rtd_theme"  # 'alabaster'

html_theme_path = [sphinx_rtd_theme.get_html_theme_path()]
# html_static_path = ['_static']



import torch
import torch_optimizer as optim

from finol.utils import *

optimizer_dict = {
    'Adadelta': torch.optim.Adadelta,
    'Adagrad': torch.optim.Adagrad,
    'Adam': torch.optim.Adam,
    'AdamW': torch.optim.AdamW,
    # 'SparseAdam': torch.optim.SparseAdam,
    'Adamax': torch.optim.Adamax,
    'ASGD': torch.optim.ASGD,
    'SGD': torch.optim.SGD,
    'RAdam': torch.optim.RAdam,
    'Rprop': torch.optim.Rprop,
    'RMSprop': torch.optim.RMSprop,
    'NAdam': torch.optim.NAdam,
    # 'LBFGS': torch.optim.LBFGS,
    'A2GradExp': optim.A2GradExp,
    'A2GradInc': optim.A2GradInc,
    'A2GradUni': optim.A2GradUni,
    'AccSGD': optim.AccSGD,
    'AdaBelief': optim.AdaBelief,
    'AdaBound': optim.AdaBound,
    'AdaMod': optim.AdaMod,
    'Adafactor': optim.Adafactor,
    # 'Adahessian': optim.Adahessian,
    'AdamP': optim.AdamP,
    'AggMo': optim.AggMo,
    'Apollo': optim.Apollo,
    'DiffGrad': optim.DiffGrad,
    'LARS': optim.LARS,
    'Lamb': optim.Lamb,
    'MADGRAD': optim.MADGRAD,
    'NovoGrad': optim.NovoGrad,
    'PID': optim.PID,
    'QHAdam': optim.QHAdam,
    'QHM': optim.QHM,
    'Ranger': optim.Ranger,
    'RangerQH': optim.RangerQH,
    'RangerVA': optim.RangerVA,
    'SGDP': optim.SGDP,
    'SGDW': optim.SGDW,
    'SWATS': optim.SWATS,
    # 'Shampoo': optim.Shampoo,
    'Yogi': optim.Yogi
}


class OptimizerSelector:
    def __init__(self, model):
        config = load_config()
        # Optimization Layer Configuration
        self.LEARNING_RATE = config["LEARNING_RATE"]
        self.OPTIMIZER_NAME = config["OPTIMIZER_NAME"]

        self.model = model

    def select_optimizer(self,  sampled_lr=None, sampled_optimizer=None):
        if sampled_lr!= None:
            self.LEARNING_RATE = sampled_lr
        if sampled_optimizer!= None:
            self.OPTIMIZER_NAME = sampled_optimizer

        optimizer_cls = optimizer_dict.get(self.OPTIMIZER_NAME, None)
        if optimizer_cls is None:
            raise ValueError(f"Invalid optimizer name: {self.OPTIMIZER_NAME}. Supported optimizers are: {optimizer_dict}")

        optimizer = optimizer_cls(self.model.parameters(), lr=self.LEARNING_RATE)
        return optimizer


from finol.evaluation_layer.EconomicDistiller import EconomicDistiller
from finol.evaluation_layer.MetricCaculator import MetricCaculator
from finol.evaluation_layer.BenchmarkLoader import BenchmarkLoader
from finol.utils import ROOT_PATH, load_config, send_message_dingding


class ModelEvaluator:
    def __init__(self, load_dataset_output, train_model_output):
        self.config = load_config()
        self.load_dataset_output = load_dataset_output
        self.train_model_output = train_model_output

    def evaluate_model(self):
        economic_distiller_caculate_metric_output = None
        if self.config["INTERPRETABLE_ANALYSIS_CONFIG"]["INCLUDE_INTERPRETABILITY_ANALYSIS"] or self.config["INTERPRETABLE_ANALYSIS_CONFIG"]["INCLUDE_ECONOMIC_DISTILLATION"]:
            economic_distiller_caculate_metric_output = EconomicDistiller(self.load_dataset_output, self.train_model_output).economic_distillation()
        caculate_metric_output = MetricCaculator(self.load_dataset_output, self.train_model_output).caculate_metric()
        load_benchmark_output = BenchmarkLoader(caculate_metric_output, economic_distiller_caculate_metric_output).load_benchmark()

        send_message_dingding(load_benchmark_output)
        return load_benchmark_output


if __name__ == "__main__":
    from finol.data_layer.DatasetLoader import DatasetLoader
    load_dataset_output = DatasetLoader().load_dataset()
    train_model_output = {
        # "logdir": ROOT_PATH[:-5] + "/logdir/" + DATASET_NAME + "-" + MODEL_NAME
        "logdir": ROOT_PATH[:-5] + "/logdir/2024-07-16_18-55-57"
    }
    evaluate_model_output = ModelEvaluator(load_dataset_output, train_model_output).evaluate_model()


import time
import torch
import numpy as np
import pandas as pd

from finol.utils import load_config, actual_portfolio_selection


class MetricCaculator:
    def __init__(self, load_dataset_output=None, train_model_output=None, mode="normal"):
        self.config = load_config()
        self.load_dataset_output = load_dataset_output
        self.train_model_output = train_model_output
        self.mode = mode

        if mode == "normal":
            self.logdir = self.train_model_output["logdir"]
        else:
            self.logdir = None

    def caculate_CW(self):
        # Calculate the Cumulative Wealth (CW)
        self.CW = self.total_returns[-1]

    def caculate_APY(self):
        # Calculate the Annual Percentage Yield (APY)
        num_years = self.num_trading_periods / self.config["NUM_DAYS_PER_YEAR"][self.config["DATASET_NAME"]]  # Number of years
        self.APY = pow(self.total_returns[-1], 1 / num_years) - 1  # APY formula

    def caculate_SR(self):
        # Calculate the Sharpe ratio (SR)
        self.sigma = np.std(self.daily_returns, ddof=1) * pow(self.config["NUM_DAYS_PER_YEAR"][self.config["DATASET_NAME"]], 0.5)  # Standard deviation of daily returns (annualized)
        self.SR = (self.APY - 0.04) / self.sigma  # Sharpe ratio formula, risk-free rate of return, reference CORN in 2011 is 0.04, reference LOAD in 2019 is 0

    def caculate_MDD(self):
        # Calculate the Maximum DrawDown (MDD)
        self.mddvec = []
        for i in range(self.num_trading_periods):
            j = i + 1
            temp = self.total_returns[:j]
            ddval = (max(temp) - temp[i]) / max(temp)  # Drawdown value for each trading period
            self.mddvec.append(ddval)
        self.MDD = max(self.mddvec)  # Maximum drawdown value

    def caculate_ATO(self):
        self.ATO = sum(self.daily_turno_list) / (2 * (self.num_trading_periods))

    def caculate_TCW(self):
        df = pd.DataFrame(columns=["Rate", "TCW"])

        for tc_rate in np.arange(0, self.config["METRIC_CONFIG"]["PRACTICAL_METRICS"]["TRANSACTIOS_COSTS_RATE"] + self.config["METRIC_CONFIG"]["PRACTICAL_METRICS"]["TRANSACTIOS_COSTS_RATE_INTERVAL"], self.config["METRIC_CONFIG"]["PRACTICAL_METRICS"]["TRANSACTIOS_COSTS_RATE_INTERVAL"]):
            portfolio_o = np.zeros_like(self.portfolios[0, :])

            self.daily_turno_list = []
            self.daily_return_list = []

            for i in range(self.num_trading_periods):
                label = self.labels[i, :]
                portfolio = self.portfolios[i, :]

                daily_turno = (abs(portfolio_o - portfolio).sum())
                self.daily_turno_list.append(daily_turno)

                daily_return = np.sum(label * portfolio) * (1 - ((tc_rate / 2) * daily_turno))
                portfolio_o = portfolio * label / daily_return

                self.daily_return_list.append(daily_return)

            df.loc[len(df)] = [tc_rate, np.cumprod(self.daily_return_list)[-1]]

        self.TCW = df["TCW"]

    def caculate_DCW(self):
        self.daily_returns = np.sum(self.labels * self.portfolios, axis=1)
        self.total_returns = np.cumprod(self.daily_returns)
        self.DCW = self.total_returns

    def caculate_portfolios(self):
        self.model = torch.load(self.logdir + "/best_model_" + self.config["DATASET_NAME"] + ".pt").to(self.config["DEVICE"])
        # self.model = self.train_model_output["best_model"]
        self.model.eval()
        self.test_loader = self.load_dataset_output["test_loader"]
        self.num_trading_periods = self.load_dataset_output["NUM_TEST_PERIODS"]  # Total number of trading periods
        self.NUM_ASSETS = self.load_dataset_output["NUM_ASSETS"]

        self.portfolios = torch.zeros((self.num_trading_periods, self.NUM_ASSETS))
        self.labels = torch.zeros((self.num_trading_periods, self.NUM_ASSETS))
        start_time = time.time()
        for i, data in enumerate(self.test_loader, 0):
            x_data, label = data
            self.labels[i, :] = label

            final_scores = self.model(x_data.float())
            portfolio = actual_portfolio_selection(final_scores)
            self.portfolios[i, :] = portfolio

        self.RT = time.time() - start_time

        self.portfolios = self.portfolios.cpu().detach().numpy()
        self.labels = self.labels.cpu().detach().numpy()

    def caculate_metric(self, portfolios=None, labels=None, runtime=None):
        if self.mode == "normal":
            self.caculate_portfolios()
        elif self.mode == "ed":
            self.portfolios, self.labels = portfolios.cpu().detach().numpy(), labels.cpu().detach().numpy()
            self.num_trading_periods, self.NUM_ASSETS = portfolios.shape
            self.RT = runtime

        # Profit Metrics
        self.caculate_DCW()
        self.caculate_CW()
        self.caculate_APY()
        self.caculate_SR()
        # Risk Metrics
        self.caculate_MDD()
        # Practical Metrics
        self.caculate_TCW()
        self.caculate_ATO()

        caculate_metric_output = {
            "logdir": self.logdir,
            "CW": self.CW,
            "DCW": self.DCW,
            "APY": self.APY,
            "SR": self.SR,
            "VR": self.sigma,
            "MDD": self.MDD,
            "DDD": self.mddvec,
            "ATO": self.ATO,
            "TCW": self.TCW,
            "RT": self.RT,
        }
        return caculate_metric_output


import torch
import torch.nn as nn
import torch.nn.functional as F

from einops import rearrange
from finol.data_layer.ScalerSelector import ScalerSelector
from finol.utils import load_config


class Transformer(nn.Module):
    def __init__(self, model_args, model_params):
        super().__init__()
        self.config = load_config()
        self.model_args = model_args
        self.model_params = model_params

        self.token_emb = nn.Linear(model_args["NUM_FEATURES_ORIGINAL"], model_params["DIM_EMBEDDING"])
        self.pos_emb = nn.Embedding(model_args["WINDOW_SIZE"], model_params["DIM_EMBEDDING"])
        self.transformer_encoder = nn.TransformerEncoderLayer(model_params["DIM_EMBEDDING"], nhead=model_params["NUM_HEADS"], dim_feedforward=model_params["DIM_FEEDFORWARD"], batch_first=True)
        self.dropout = nn.Dropout(p=model_params["DROPOUT"])
        self.fc = nn.Linear(model_params["DIM_EMBEDDING"], 1)

    def forward(self, x):
        batch_size, num_assets, num_features_augmented = x.shape
        DEVICE = x.device

        """Input Transformation"""
        x = x.view(batch_size, num_assets, self.model_args["WINDOW_SIZE"], self.model_args["NUM_FEATURES_ORIGINAL"])
        x = rearrange(x, "b m n d -> (b m) n d")
        if self.config["SCALER"].startswith("Window"):
            x = ScalerSelector().window_normalize(x)

        """Temporal Representation Extraction"""
        x = self.token_emb(x)  # [batch_size * num_assets, window_size, num_features_original] -> [batch_size * num_assets, window_size, DIM_EMBEDDING]
        pos_emb = self.pos_emb(torch.arange(self.model_args["WINDOW_SIZE"], device=DEVICE))
        pos_emb = rearrange(pos_emb, "n d -> () n d")
        x = x + pos_emb

        output = self.transformer_encoder(x)
        output = output[:, -1, :]

        """Final Scores for Assets"""
        output = output.view(batch_size, num_assets, self.model_params["DIM_EMBEDDING"])
        output = self.dropout(output)
        output = F.relu(output)
        final_scores = self.fc(output).squeeze(-1)
        return final_scores
